#ifndef _CPLRAW_FAILABLE_HXX_
#define _CPLRAW_FAILABLE_HXX_

#ifndef _MM_NO_IOSTREAM
#include <iostream>
#endif 
#include <string.h>
#undef E_FAIL /* windows error code */

  /**
   * base class for something that does have a failed/good state. 
   */
  class Failable {

    int _iErr;
    const char *_szErr;		/* must be static	*/

   protected:

    enum StatE {
      E_OK,
      E_FAIL = 001,
      E_BAD =  002 ,
      E_NEXT = 004
    }  _eStat;

    Failable& setStat(StatE eStat, const char *sz, int i);
    Failable& setStat(const Failable &fa);

    /**
     * resets the failed state. returns !good
     */
    bool entry();


    #ifndef _MM_NO_IOSTREAM
      friend std::ostream &operator<<(std::ostream &, const Failable &);
    #endif
   public:
    bool isGood() const { return _eStat == 0; }		/**< fine 		*/
    int print(char *sz, unsigned size) const;
    int iErr() { return _iErr; }			/**< error code		*/
    const char *szErr() { return _szErr; }		/**< error text		*/

    /** 
     * last op failed, but object is uncorrupted
     */
    bool isFail() const { return (_eStat & (E_FAIL|E_BAD))!=0; }

    /** 
     * last op failed, object is corrupted
     */
    bool isBad() const { return (_eStat & E_BAD) != 0; }

    /**
     * true if failed
     * \code
     * if (!fa) do_failed();
     * \endcode
     */
    bool operator!() const { return isFail(); }

    /**
     * zero if failed
     * \code
     * if (fa) do_good();
     * \endcode
     */
    operator const void * () const { return isFail() ? 0 : this ; }

    /**
     * set 'this' failed, record a string and an integer for failure identification
     * \code
     * if (i==-1) fa.setFailed(__FILE__, __LINE__);
     * \endcode
     */
    Failable &setFailed(const char*sz, int i=0) { return setStat(E_FAIL, sz, i); }

    /**
     * set this bad, record a string and an integer for failure identification
     */
    Failable &setBad(const char*sz, int i=0) { return setStat(E_BAD, sz, i); }

    /**
     * user clear of failed/bad state
     */
    Failable &clear() { return setStat(E_OK, 0, 0); }

    Failable();
    Failable(const char *sz, int i);	/* removed default for i */
    ~Failable(); 
  };

  /** Failable with reason stack */
  class Failables:public Failable {
    int _c;
    enum { ccFa=5 } ;
    Failable _mpfa[ccFa];

    #ifndef _MM_NO_IOSTREAM
      friend std::ostream &operator<<(std::ostream &, const Failables &);
    #endif
   public:

    /** if pfa, initialize history with *pfa */
    Failables &set(StatE e, const char*sz, int i, const Failable *pfa);

    /** if pfa, initialize history with *pfa */
    Failables &set(StatE e, const char*sz, int i, const Failables *pfa);


    Failables &setFailed(const char*sz, int i, const Failable *pfa) { return set(E_FAIL, sz, i, pfa); }
    Failables &setBad(const char*sz, int i, const Failable *pfa) { return set(E_BAD, sz, i, pfa); }

    Failables &setFailed(const char*sz, int i=0, const Failables *pfa=0) { return set(E_FAIL, sz, i, pfa); }
    Failables &setBad(const char*sz, int i=0, const Failables *pfa=0) { return set(E_BAD, sz, i, pfa); }

    /**
     * user clear of failed/bad state
     */
    Failables &clear();

    Failables();
    ~Failables();
  };

#endif
