#ifndef _TINA_TINA_HXX
#define _TINA_TINA_HXX

#include <assert.h>
#include <list>
#include <deque>
#include <algorithm>
#include "cplRaw/flags.hxx"
#include "cplRaw/dbg.hxx"
#include "umi/umi_task.hxx"
#include "umipro_tina.hxx"
#include "tina/tico_pub.hxx"
#include <stdint.h>
#undef min
#undef max

/**
 * \file 
 * ines Timing Analyzer (Tina) interface 
 */


/** namespace for the ines Timing Analyzer (Tina) interface */
namespace INdepTina100 {
  namespace Umi=INdepUmi100 ;



  typedef double Freq;
  typedef Umi::Time Time;
  typedef std::pair<Counter, Timer> Edge;
  typedef std::pair<Timer, Timer> Pulse;	/* high / low */


  /**
   * Timing Analyzer Mode
   */
  class  TinaMode:public Failable, public DebugLog {
    const int _cFifo;		/* count of fifos/capture_units used (1-2)	*/
    const TiCore *_ptc;
   protected:
    TinaConfig _tc;		/* server config request	*/
    unsigned _cBlock;		/* sample count maximum, 0=unlimited  */
    Timer _tmTmo;
    Timer _tmLim;
    union {
      /* current epoch per channel */
      Timer _tmEpoch;		
      Timer _mptmEpoch[2];
    };
   private:
    Umi::Time _tiTmo;
   protected:

    int timeout(Timer tmCur, bool fReset);

    const TiCore *ptc() const { return _ptc; }	/* set by init() */


   public:
    const TinaConfig &tc() { return _tc; }
    
    

    unsigned cBlock() const { return _cBlock; }
    int cFifo() const { return _cFifo; }
    int cCapt() const { return _cFifo; }	/* capture units required */
    Umi::Time tiTmo() { return _tiTmo; }

    /* per ECU flags */

    bool fEdgeRise(int i) const { return i; }   /* 1 = rising, 0= falling */
    bool fModeFilt(int i) const { return 0; }   /* 1 = filtered, 0 = raw  */

    virtual void sample( std::streambuf *psb, int iFifo ) = 0;
    virtual void init( const TiCore *ptc );
    virtual bool isDone() const =0;
    virtual unsigned cSamp() const =0;

    TinaMode(int cFifo, unsigned cBlock, const Umi::Time &tiTmo);
    TinaMode():_cFifo(0) { assert(0); }
    virtual ~TinaMode();
  };

  /**
   * Frequency/Period Average measurement mode
   */
  class  ModeAvg : public TinaMode {
    /* constructor arguments */
    unsigned _cScale;		/* requested total scale   */
    unsigned _cRep;		/* scale (software scaler) */
    unsigned _iRep;		

    Timer _tiPrev;
    std::deque<Timer> _dqTimes;

   private:
    void init( const TiCore *ptc );

    /** initialize from hardware buffer */
    void sample( std::streambuf *psb, int iFifo );

   public:
    /**
     * true when the requested count of measurements (defined by the cBlock constructor argument) is available
     */
    bool isDone() const { return cBlock() && _dqTimes.size() > cBlock(); }

    /**
     * current count of measurements available for read via rdFreq(), rdPeriod()
     */
    unsigned cMeas() const { return _dqTimes.size() > 1 ? _dqTimes.size() -1 : 0; }
    unsigned cSamp() const { return cMeas() ; }

    /**
     * \name Measurement returns
     * These functions fetch the current (least recent) measurement from the input queue.
     * They return 0.0 if called erroneously (i.e. with no measurement available).
     */
    /**@{*/
    double rdFreq();		/**< frequency average */
    double rdPeriod();		/**< period average */
    /**@}*/

    /**
     * prepare next block
     */
    void clear();
    

     /**
      * Measurement Mode constructor
      * \arg cAverage averaging factor, must be > 0
      * \arg cBlock block size until isDone() is true, 0=unlimited  
      * \arg tiTmo inactivity (no input edges) timeout 
      */
    ModeAvg(unsigned cAverage, unsigned cBlock=1, const Umi::Time &tiTmo=Umi::Time(1));

     /**
      * Measurement Mode destructor,
      * releases all resources bound to this measurement mode
      */
    ~ModeAvg();
  };


  class ModeDual:public TinaMode {
    typedef std::deque<Timer> Segment;	/* segment of continous events */
    std::deque<Segment> _mpdqSeg[2];
     
   protected:

    std::deque<Timer>& dqTimes(int i) { return _mpdqSeg[i].back() /*._dqTimes*/; }
    const std::deque<Timer>& dqTimes(int i) const { return _mpdqSeg[i].back()/*._dqTimes*/; }

    enum { LE=0, TE=1 };	/* leading/trailing edge */
    const unsigned _cExtraLE;		/* pulse mode needs terminating LE */


    /** initialize from hardware buffer */
    void sample( std::streambuf *psb, int iFifo );
    void init( const TiCore *ptc );
    //unsigned cSeg() const { return _dqSeg.size(); }
    //unsigned cSegRq() const { return 1; }

    unsigned sizeLE() const { return dqTimes(LE).size()<_cExtraLE ? 0:dqTimes(LE).size()-_cExtraLE; }
    unsigned sizeTE() const { return dqTimes(TE).size(); }
   public:
    unsigned cSamp() const { return std::min(sizeLE(), sizeTE()); }
    bool isDone() const { return cBlock() && cSamp() >= cBlock(); }

   protected:
    ModeDual(unsigned cSampRq, const Umi::Time &tiTmo, unsigned cExtraLE):TinaMode(2, cSampRq, tiTmo),_cExtraLE(cExtraLE) {  
      _mpdqSeg[LE].push_back(Segment());
      _mpdqSeg[TE].push_back(Segment());
    }
    ~ModeDual() {  }
  };

  /**
   * Delay A to B channel
   */
  class ModeDelay:public ModeDual {
  };

  /**
   * Pulse Width
   */
  class ModePulseWidth:public ModeDual {
    int _eMode; 		/* 'H'=high width, 'L'=low width */
   public:
    double rdPulseWidth();

 
    //Failable &setMode(int eMode) ;
    //Failable &setTimeout(const Time &ti) ;
    ModePulseWidth(int eMode='H', unsigned cSampRq=1, const Umi::Time &tiTmo=Umi::Time(1));
    ~ModePulseWidth();
  };

  /**
   * Pulse Width
   * REMIND: implement sampled mode (one capt per sample clock)
   * REMIND: implement pulse mode (low/high pair, ratio)
   */
  class ModePulse:public ModeDual {
    int _eMode; 		/* 'H'=high width, 'L'=low width */
    std::pair<double, double> _pardPulse;
   public:

    std::pair<double, double> pardPulse();
    double rdFirst() const { return _pardPulse.first; }
    double rdSecond() const { return _pardPulse.second; }
 
    //Failable &setMode(int eMode) ;
    //Failable &setTimeout(const Time &ti) ;
    ModePulse(int eMode='H', unsigned cSampRq=1, const Umi::Time &tiTmo=Umi::Time(1));
    ~ModePulse();
  };




  /**
   * Edge counts sampled on interval or external signal
   *
   * REMIND up/down mode possible (=> two counter difference) ?
   */
  class ModeTotalCount:public TinaMode, public std::list<Edge> {
    int _eMode; 	/* R=rising, F=falling edge */
    Counter _ctEpoch;	/* event counter epochs */

    /** initialize from hardware buffer */
    void sample( std::streambuf *psb, int iFifo );
    void init( const TiCore *ptc );

    std::deque<Counter> _dqCounts;

   public:
    bool isDone() const { return cBlock() && _dqCounts.size() >= cBlock(); }
    unsigned cSamp() const { return _dqCounts.size(); }
    Counter ctCurrent() { return begin()->first; }

    Counter count();

    int setEdge(int eMode) { return (eMode == 'R' || eMode == 'F') ? _eMode=eMode, 0 : -1 ; }
    ModeTotalCount(int eMode='R', unsigned cSampRq=1, const Umi::Time &tiTmo=Umi::Time(1));
    ~ModeTotalCount();
  };


  /* sequence of pulse widths */
  class ModePulseWidthSequence:public TinaMode, public std::list<Time> {
  };

  class ModePulseWidthSingle: public ModePulseWidthSequence {
    Time tiCurrent() { return *begin(); }
  };

  /*
   * Pulse Period
   */

  /* samples of most-recent pulse seen */
  class TimoPulsePeriodSamples:public TinaMode, public std::list<Pulse> {
    int _eMode; 	/* 1=leading, -1 = trailing edge */
   public:

    /** initialize from hardware buffer */
    void sample( const char mpch[6] ) ;
 
    Failable &setMode(int eMode) ;
    Failable &setTimeout(const Time &ti) ;
    TimoPulsePeriodSamples() { }
    ~TimoPulsePeriodSamples() { }
  };

  /* sequence of pulse widths */
  class TimoPulsePeriodSequence:public TinaMode, public std::list<Pulse> {
  };

  class TimoPulsePeriodSingle: public TimoPulsePeriodSequence {
    Pulse puCurrent() { return *begin(); }
  };

  /*
   * Period/Frequency Average
   * for increased precision, higher frequencies, higher troughput
   */
  class TimoFreqAvgSequence:public TinaMode, public std::list<Freq> {
  };

  /* Start/Stop */
  class TimoStartSglStop:public TinaMode { };
  class TimoStartMltStop:public TinaMode { };

  /**
   * Pulse Generator Mode
   */
  class PulseLine : public Time {
    bool _fChip;
   public:
    bool fChip() const { return _fChip; }
    PulseLine(const Time &ti, bool f):Time(ti), _fChip(f) { }
    ~PulseLine() { }
  };

  class  PulseMode:public Failable, public std::deque<PulseLine> {
   public:
    virtual int cPulse( ) { return size(); }
    virtual void pulse( char mpch[6], const TiCore *ptc );
    PulseMode();
    ~PulseMode();
  };
  extern Time ti(double rd);

  class PumoFreq:public PulseMode {
   public:


    void init(double rdWidthLow, double rdWidthHigh) {
      push_front(PulseLine(ti(rdWidthLow), 0));
      push_front(PulseLine(ti(rdWidthHigh), 1));
    }

    PumoFreq(double rdFreq, double rdDuty=0.5) {
      if (rdDuty > 1.0) rdDuty=1.0;
      if (rdDuty < 0.0) rdDuty=0.0;
      init(rdDuty/rdFreq,(1.0-rdDuty)/rdFreq);
    }

    //PumoFreq(double rdWidthLow, double rdWidthHigh) { init(rdWidthLow, rdWidthHigh); }



    ~PumoFreq() { }
  };

  

  class TinaPulseConfigClientFunc:public Umi::TaskFunc::Func {
    int _cnServer;
    PulseMode *_ppm;
    const TiCore *_ptc;

    int cnServer() const { return _cnServer; }

   public:
    int func( int eState ) {
      switch (eState) {
        case 0: {		/* init */
          return setup(1, fCanput(cnServer()));
	} break;
	case 1: {
	  /* write data rq */
	  new(pmu(cnServer())->puc()) TinaPulseConfig(); 
	  while(_ppm->cPulse()) {
	    char mpch[6];
	    _ppm->pulse(mpch, _ptc);
	    for (unsigned i=0;  i < 6; i++) {
	      assert(pmu(cnServer())->pup()->canput());
	      pmu(cnServer())->pup()->sputc(mpch[i]); 
	    }
	  }
	  return setup(2, fCanget(cnServer()), fSendup(cnServer()));
	} break ;
	case 2: {
	  /* poke rp */
	  TinaResp* ptr=pTinaResp(pmd(cnServer())->puc()); 
	  assert(ptr);
	  ptr+=0;
	  return setup(-1, 0, fReleasedown(cnServer()));
	} break ;
	default: assert(0);
      }
      return 0;
    }
    TinaPulseConfigClientFunc(int cnServer, PulseMode *ppm, const TiCore *ptc):_cnServer(cnServer),  _ppm(ppm), _ptc(ptc)  { }
    ~TinaPulseConfigClientFunc() { }
  };


  /**
   * Tina Protocol Plexer
   */
  class TinaPlex:public Umi::TaskPlex,  public Failable {
    Umi::ChanUser *_pucIo;
    Umi::Channel *_puc;
    const TiCore *_ptc;

    bool _fRun;

    void done(Umi::Task *pta) { 	/* TaskUser virtual */
      _fRun=0; 
    }





    void run(Umi::TaskFunc::Func *ptf) {
      assert(!_fRun); 
      _fRun=1; 

      TaskFunc::run(ptf);
      sigUp(0, _puc, 0);
      runio(_pucIo); 
      assert(!_fRun); 
    }

   public:

    /**
     * configure timing analzer for measurement mode specified by ptm
     */
    int callTinaConfig(TinaMode *ptm);

    /**
     * sample data for measurement mode specified by ptm
     */
    int callTinaSample(TinaMode *ptm);

    int callPulseConfig(PumoFreq *ptm) { 
      TinaPulseConfigClientFunc *ptf=new TinaPulseConfigClientFunc(0, ptm, _ptc);	/* deleted by task code ! */
      run(ptf);
      return ptm->isGood();
    }

    
    TinaPlex(Umi::Channel *puc, Umi::ChanUser *pucIo):Umi::TaskPlex(puc), 
      _pucIo(pucIo), 
      _puc(puc), 
      _ptc(&tc2n5),
      _fRun(false)
    { }
  };

}

#endif

