#ifndef _UMI_UMI_USER_HXX
#define _UMI_UMI_USER_HXX

/**
 * \file umi_user.hxx Universal Message Interface - User C++ part
 *
 * %LEGAL%
 */
#include "umic/umi.hxx"
#include "umic/umi_driver.hxx"
#include "cplRaw/failable.hxx"

#ifndef EOF 
  #define EOF (-1)
#endif


namespace INdepUmi100 {

  /**
   * Process local Umi channel
   */
  class Lock;
  class UmiChanLocal : public UmiChannel {
    UmiChanLocal *_pucPeer;	/**< peer channel */

    /** create connected */
    void finiPool(Link *pli);
    friend bool sendMessage(int iCtx, UmiChanLocal *pucSend, UmiChanLocal *pucRecv);
   public:
    UmiChanLocal *pucPeer() { return _pucPeer; }

    void sigGet(int iCtx);	/**< new message ready for pickup at the upstream queue. */
    void sigPut(int iCtx); 	/**< room at the downstream queue to accommodate at least one downstream message.  */

    void initPeer(UmiChanLocal *pucPeer) { _pucPeer=pucPeer; }

    UmiChanLocal(const UmiProto *pup);
    ~UmiChanLocal();
  };

  class LocalChanPair:public Failable {
    UmiChanLocal *_pucA;	/* local channel pair (TinaPlex) */
    UmiChanLocal *_pucB;
   public:
    UmiChanLocal *pucA() { return _pucA; }
    UmiChanLocal *pucB() { return _pucB; }

    LocalChanPair(): _pucA(0), _pucB(0) { }

    Failable &init(const Proto *pup) {
      /* create and link two local channels */
      _pucA=new UmiChanLocal(pup);
      _pucB=new UmiChanLocal(pup);
      _pucA->initPeer(_pucB);
      _pucB->initPeer(_pucA);
      return *this;
    }

    void fini() {
      delete _pucB;
      delete _pucA;
      _pucA=0;
      _pucB=0;
    }
    ~LocalChanPair() { fini(); }
  };

  class LocalChannelPair:public LocalChanPair {
    Plexer *_pux;
    Plexer *_puxClient;
    int _ePort;
    int _ePortClient;
   public:
    UmiChanLocal *pucClient() { return pucB(); }
    Plexer *puxServer() const { return _pux; }
    Plexer *puxClient() const { return _puxClient; }

    LocalChannelPair():_pux(0), _puxClient(0), _ePort(-1), _ePortClient(-1)  { }

    Failable &init(Plexer *pux, const char *szPort, const Proto *pup) {
      _pux=pux;
      LocalChanPair::init(pup);

      if ((_ePort=_pux->openChannel(szPort, pucA()))<0) {
	setFailed("Plexer openChannel() failed", 'a');
      }
      return *this;
    }

    Failable &init(Plexer *pux, const char *szPort, Plexer *puxClient, const char *szPortClient, const Proto *pup) {
      init(pux, szPort, pup);
      _puxClient=puxClient;

      if ((_ePortClient=_puxClient->openChannel(szPortClient, pucB()))<0) {
	setFailed("Plexer openChannel() failed", 'b');
      }
      return *this;
    }

    void fini() {
      if (_pux && _ePort >= 0) _pux->closeChannel(_ePort);
      if (_puxClient && _ePortClient >= 0) _puxClient->closeChannel(_ePortClient);
      LocalChanPair::fini();
      _pux=0;
      _ePort=-1;
      _puxClient=0;
      _ePortClient=-1;
    }
    ~LocalChannelPair() { fini(); }
  };

#if (defined(MMSYS_linux) || defined(MMSYS_qnx))
  /* Umi Perimeter Lock */
  class Lock {
    pthread_mutex_t _muUmi;
    int _fLock;
   public:
    int isLocked() { return _fLock; }
    void lock() { pthread_mutex_lock(&_muUmi); _fLock++; }
    void unlock() { _fLock--; pthread_mutex_unlock(&_muUmi); }
    Lock() { pthread_mutex_init(&_muUmi, 0); _fLock=0; }
    ~Lock() { pthread_mutex_destroy(&_muUmi); }
  };

  class LockMutex {
    Lock *_pul;
     public :
   LockMutex(Lock *pul):_pul(pul) { if (_pul) _pul->lock(); }
   ~LockMutex() { if (_pul) _pul->unlock(); }
  };
#endif
 

  /**
   * Base class for user space Umi channels, e.g. ioctl / socket
   */
  class UmiChanUser : public UmiChannel, public Failables {
    UmiMsgBuf *_pub;
    UmiMsgBuf *_pdb;

   protected:
    void setupSend();
    void setupRecv();

    void clrPub() { _pub=0; }
    void clrPdb() { _pdb=0; }

    UmiMsgBuf *pub() const { return _pub; }
    UmiMsgBuf *pdb() const { return _pdb; }

   public:
    const UmiMsgBuf *pubDebug() const { return _pub; }		/* public for debug only */
    const UmiMsgBuf *pdbDebug() const { return _pdb; }

    /**
     * \name Transport side transfer
     *
     * These functions are called for data transmission
     * when the transport is ready to send or receive.
     * They have exclusive access to the current buffer(s), thus
     * synchronization with the channel is not required.
     */
    //@{
    virtual Failables &sigSendTransfer()=0;
    virtual Failables &sigRecvTransfer()=0;
    //@}

    /**
     * \name Transport side management
     *
     * These functions must be called after any data transfer.
     * They need exclusive (possibly locked) access to the channel's perimeter.
     */
    //@{
    Failables &sigSendDone();	
    Failables &sigRecvDone();
    //@}

    /**
     * \name Transport side signaling
     *
     * Convenience functions for data transfer if mutual exclusion is ensured by design.
     * Alternatively use the sequence: sigSendTransfer, exclusive(sigSendDone).
     */
    //@{
    Failables &sigSend() { sigSendTransfer(); sigSendDone(); return *this; }
    Failables &sigRecv() { sigRecvTransfer(); sigRecvDone(); return *this; }
    //@}

    /**
     * \name Umi side signaling
     *
     * These functions are called from the base class with the perimeter locked. and are not allowed to do any
     * time consuming task. Their implementation shall signal an external ("worker") thread which calls
     * corresponding functions when ready to send or receive.
     * 
     */
    //@{
    void sigGet(int);	/**< called with perimeter locked! */
    void sigPut(int);   /**< called with perimeter locked! */
    //@}

    //void initRecv() { setupRecv(); }

    /**
     * \name Transport side arming
     *
     * This functions are called when buffers are set up to send or receive.
     *
     * They are not allowed to do any time consuming task. 
     * Their implementation shall signal an external ("worker") thread which calls
     * corresponding functions when ready to send/receive.
     * 
     */
    //@{
    virtual void armSend()=0;	/* arm sigSend() */
    virtual void armRecv()=0;	/* arm sigRecv() */
    //@}

    void finiPool(Link *pli);


    /** create connected */
    UmiChanUser(const UmiProto *pup);
    ~UmiChanUser();
  };
  typedef UmiChanUser ChanUser;

  /* run request/response sequence, nonzero on pucIo failure */
  extern int runio(ChanUser *pucIo); 
  extern int runio(UmiChanUser *pucIo, int (*pfHook)(void *), void *pv); /* new version, break if hook returns nonzero */



  /**
   * Umi synchronous call interface.
   *
   * This interface has been called Umi in earlier versions.
   *
   * \note 'Failure' in the context of a Umi always means 'transportation' 
   * never 'endpoint' failure.
   */
  class UmiCall:public Failables {
    const UmiProto *_pup;

   public:
    const UmiProto *pup() { return _pup; }
    /*
     * transport failures
     */
    enum {
     OK,
     E_NOBUF
    };

    /**
     * send a message downstream.
     * \param puc Pointer to the control part of the message
     * \param pud Pointer to the data part of the message
     * \return (Failable&) *this
     */
    virtual Failables& putMsg(const UmiControl*puc, const UmiData*pud) = 0;

    /**
     * get a upstream message
     * \param puc Pointer to the control part buffer
     * \param pud Pointer to the data part buffer
     * \return (Failables&) *this
     */
    virtual Failables& getMsg(UmiControlBuf *puc, UmiData *pud) = 0;

    /**
     * exchange message
     * \param pucd Pointer to the downstream control part of the message
     * \param pudd Pointer to the downstream data part of the message
     * \param pucu Pointer to the upstream control part buffer
     * \param pudu Pointer to the upstream data part buffer
     * \return (Failables&) *this
     */
    virtual Failables& xchMsg(const UmiControl*pucd, const UmiData*pudd, UmiControlBuf *pucu, UmiData *pudu);

    /* flow control prevent deadlock */
    virtual bool canPut() ;


    /* data max per message */
    size_t sDataMax() { return __SZ_DATA_MAX; }		

    UmiCall(const UmiProto *puf);
    virtual ~UmiCall();
  };



  /**
   * Factory interface definition.
   *
   * Using a Factory, UMI channels can be constructed without
   * knowledge of the specific UMI in use. All channels for a
   * specific protocol must be created using the same Factory.
   * 
   * \note In order to assist factories which keep state between the
   * creation of channels, all channels must be created in increasing channel 
   * order ( 0, 1, ... ).
   * Furthermore, the Umi channel objects created by createUmi() must be deleted
   * in reverse order ( ... , 1 , 0 ).
   */
  class Factory {
   public:
    /**
     * create channel to driver specified by protocol and channel number
     * (thread-safe)
     * \param ppum Pointer to Umi*, returns the UMI created. Set to 0 if the creation did fail.
     * \param pup Pointer to the UMI protocol to be spoken on the channel.
     * \param iChan channel specification.
     * \return A Failables indicating the state of the created Umi, even if this has not been created and is not returned.
     */
    virtual Failables createUmi(UmiCall**ppum, const UmiProto *pup, int iChan) =0;

    Factory() {} 
    virtual ~Factory() {}
  };

  /** factory maker */
  class FactoryImplementation {
    const char *_szName;
   public:
    const char *szName() { return _szName; }
    virtual Factory* createFactory(const char *szMode, const char *szArg)=0;
    FactoryImplementation(const char *szName):_szName(szName) {}
    virtual ~FactoryImplementation() {}
  };

  /**
   * implementation dependent list of Factories
   * (thread-safe)
   */
  class Factories {

    /** 
     * glocal factory table, to be set up by global initializers
     */
    static FactoryImplementation* mpPff[4];

   public:
    void registerFactoryImplementation(FactoryImplementation *);
    
    /** 
     * construct Factory from internal list
     */
    Factory* createFactory(const char *szArg);

    Factories();
     ~Factories();
  };

  /**
   * implemented factories (global)
   */
  extern Factories fact;
}
#endif


