#ifndef _UMI_C_H_
#define _UMI_C_H_

/**
 * \file umi_c.h 
 * Universal Message Interface - C core interface.
 * Enables access to the C++ classes via a C-type interface. 
 * Required in particular by the Linux kernel as it does not allow C++ coding.
 */

#include <stddef.h>
#include "umi.h"
#include "umiad.h"


  struct UmiIter {
     char _mpch[8*8];
  };

  /** memory block descriptor */
  struct UmiMemInfo {
    size_t _s;
    void * _p;
  };

  struct UmiCard ;
  struct UmiMod ;
  struct UmiDriver ;
  struct UmiChanKern ;
  struct UmiProto ;
  struct UmiControl ;
  struct UmiMsgDownBuf ;
  struct UmiMsgUpBuf ;

  /**
   * identfies to the specific runtime environment to
   * the UMI driver module.
   */
  struct UmiEnvC {
    void (*_pFtiNextTick)(struct UmiTimespec *);
    void (*_pFsigGet)(int, void*); /**< Environment function to be called if a message is available     */
    void (*_pFsigPut)(int, void*); /**< Environment function to be called if a message can be processed */
  };



  /**
   * per driver protocol and channel count
   */
  struct UmiDriverDesc {
    unsigned _cChan;			/**< channels per driver		*/
    const struct UmiProto *_pup;	/**< driver protocol 			*/

    /* unused obsolete fields for compatibilty */
    //size_t _xxx__sAlloc;
    //void *(*_xxx___pFnew_UmiDriver)(struct UmiMod*, struct UmiMemInfo*pum, const struct UmiCard*);
  };

  /**
   * driver constructor must match the corresponding UmiDriverDesc
   */
  struct UmiDriverObject {
    size_t _sAlloc;			/**< driver module allocation size	*/

    /**
     * pointer to the constructor (C-Function).
     * The constructor function initializes the memory space given by pum.
     */
    void *(*_pFnew_UmiDriver)(struct UmiMod*, struct UmiMemInfo*pum, const struct UmiCard*);
  };

  /**
   * identfies a UMI driver module to the specific runtime environment
   */
  struct UmiMod {
    int _cChan;			/**< channel per card		*/
    int _tps;			/**< timer ticks per second	*/

    /**
     * pointer to the constructor (C-function)
     *
     */
    void *(*_pFnew_UmiCard)(struct UmiMod*, const struct UmiEnvC*, struct UmiMemInfo*, int, void* pArg, long lAddr);

    /**
     * pointer to the allocator (C-function)
     */
    int (*_pFgetMem_UmiCard)(struct UmiMod*, struct UmiMemInfo*, int);

    /***
     *** new interface 
     ***/
    unsigned _cDriver;			/**< driver count		*/
    struct UmiDriverDesc *_puDd;	/**< driver descriptors		*/

    size_t _sAlloc;			/**< card allocation size	*/

    /**
     * pointer to the card constructor (C-function)
     */
    void *(*_pFnewUmiCardEx)(struct UmiMod*pum, struct UmiMemInfo*pmi, long lAddr, void *pArg, int fIntr, const struct UmiEnvC*);

    /**
     * pointer to the constructor of this structure. Must be called exactly once
     * before \em any other field is used.
     */
    void (*_pFct)(void);

    /**
     * list of implemented adaptors 
     */
    const UmiAdapt **_mppad;	
  };

  #define UmiMod_init(pum) ((*(pum)->_pFct)())


 #ifndef _NO_UMI_C_DECL	
  /**
   * get the description of driver i of the module
   * \ret pointer to the UmiDriverDesc, or 0 if i is out of range
   */
  extern const struct UmiDriverDesc *UmiMod_getDriverDesc(const struct UmiMod *, unsigned iDriver);
  #define UmiDriverDesc_getChanCount(pdd) ((pdd)->_cChan)
  #define UmiMod_getDriverCount(pum) ((pum)->_cDriver)


  extern int UmiCard_timer( struct UmiCard*, int, const struct timespec *);
  extern int UmiCard_interrupt( struct UmiCard*, int );

  /*
   * generic methods
   */
  extern void UmiCard_startIterDrivers( struct UmiCard*, struct UmiIter *);  /**< iterator */
  extern struct UmiDriver * UmiCard_iterateDrivers( struct UmiCard*, struct UmiIter *); /**< next driver of card */
  extern const char * UmiDriver_szName( struct UmiDriver *);	/**< driver name	*/
  extern char * UmiDriver_print( struct UmiDriver *, char *, int );
  extern int UmiDriver_init( struct UmiDriver *);
  extern void UmiDriver_fini( struct UmiDriver *);

  extern void UmiDriver_startIterChannels( struct UmiDriver*, struct UmiIter *); 
  //extern struct UmiChanKern* UmiDriver_iterateChannels( struct UmiDriver*, struct UmiIter *); /**< next channel of driver */
  extern struct UmiChannel* UmiDriver_iterateChannels( struct UmiDriver*, struct UmiIter *); /**< next channel of driver */
  extern int UmiChanKern_num( struct UmiChanKern *);
  extern const char* UmiChanKern_szProto( struct UmiChanKern *);
  extern void UmiChanKern_setHandle( struct UmiChanKern *, void *p);

  /*
   * channel upstream 
   */
  extern int UmiChanKern_canget( struct UmiChanKern *);
  extern struct UmiMsgUpBuf *UmiChanKern_getUpBuf( struct UmiChanKern *);			/**< next to get		*/
  extern void UmiChanKern_releaseUpBuf( struct UmiChanKern *, int iCtx, struct UmiMsgUpBuf *);	/**< last got		*/

  /*
   * channel downstream 
   */
  extern int UmiChanKern_canput( struct UmiChanKern *);
  extern struct UmiMsgDownBuf *UmiChanKern_allocDownBuf( struct UmiChanKern *);	/**< next to put		*/
  extern void UmiChanKern_putDownBuf( struct UmiChanKern *, int iCtx, struct UmiMsgDownBuf *); /**< last got		*/

  extern size_t UmiMsgUpBuf_sizeControl(struct UmiMsgUpBuf *);
  extern struct UmiControlStruct * UmiMsgUpBuf_pControl(struct UmiMsgUpBuf *);
  extern size_t UmiMsgUpBuf_sizeData(struct UmiMsgUpBuf *);
  extern char * UmiMsgUpBuf_pData(struct UmiMsgUpBuf *);

  extern size_t UmiMsgDownBuf_sizeControl(struct UmiMsgDownBuf *);
  extern size_t UmiMsgDownBuf_sizeDataMax(struct UmiMsgDownBuf *);
  extern void UmiMsgDownBuf_setDataLim(struct UmiMsgDownBuf *, size_t);
  extern struct UmiControlStruct * UmiMsgDownBuf_pControl(struct UmiMsgDownBuf *);

  extern char * UmiMsgDownBuf_pData(struct UmiMsgDownBuf *);

  #define UmiMod_cChan(pum) (pum->_cChan)
  #define UmiCard_new(pum, pue, a, b, c, d) ((*pum->_pFnew_UmiCard)((pum), (pue), (a), (b), (c), (d)))
  #define UmiCard_getMem(pum, a, b) ((*pum->_pFgetMem_UmiCard)(pum, (a), (b)))
  extern char * UmiCard_print( struct UmiCard *, char *, int );

  /**
   * marshal control (required for Linux/USB)
   * \return count of valid characters in *pch, 0 if failed
   */
  extern size_t UmiProto_marshal(const struct UmiProto*pup, char *pch, size_t, struct UmiControlStruct *puc);

  /**
   * unmarshal control (required for Linux/USB)
   * \return 0 on success, nonzero otherwise
   */
  extern size_t UmiProto_unmarsh(const struct UmiProto *pup, struct UmiControlStruct *puc, char *pch, size_t s );

  /**
   * check protocol
   * \return true if protocol is compatible, false otherwise
   */
  extern int UmiProto_match(const struct UmiProto *pupServer, char *pchClient, size_t sClient );
  extern const char * UmiProto_szProto(const struct UmiProto *pup);
  extern const char * UmiProto_szName(const struct UmiProto *pup);

 #endif

#endif
