#ifndef _UMI_UMI_DRIVER_HXX
#define _UMI_UMI_DRIVER_HXX

/**
 * \file umi_driver.hxx
 * Universal Message Interface - Driver C++ part
 */

#include "umi.hxx"
#include "cplRaw/list.hxx"
#include "cplRaw/failable.hxx"


namespace INdepUmi100 {

  /**
   * unified (up/down) message kernel rep
   */
  class UmiMsg:public Link {
    UmiControl *_puc;
    UmiBuf *_pub;
   public:
    /** message control part	*/
    UmiControl *puc() const { return _puc; }
    /** message data part	*/
    UmiBuf *pug() const { return _pub; }
    UmiBuf *pup() const { return _pub; }
    UmiBuf *pub() const { return _pub; }

    void init(UmiControl *puc, UmiBuf *pub) { _puc=puc; _pub=pub; }

    UmiMsg():_puc(0), _pub(0) {}
    ~UmiMsg() {}
  };

 #if (1)	/* compatibility with previous versions */
  typedef UmiMsg UmiMsgDown;
  typedef UmiMsg UmiMsgUp;
  typedef const UmiMsg *Pmd;
  typedef const UmiMsg *Pmu;
 #endif

  /**
   * self-contained message copy
   */
  class UmiMsgBuf:public UmiMsg, public UmiBuf, public UmiControlBuf { 
    char _mpchData[ __SZ_DATA_MAX ];
   public:
    #ifndef _MM_NO_IOSTREAM
     friend std::ostream &operator<<(std::ostream &os, UmiMsgBuf &ud);
    #endif

    /*
     * copy in user messages
     */
    UmiControl& operator=(const UmiControl &uc) { return UmiControlBuf::operator=(uc); }

    UmiMsgBuf() ;
    ~UmiMsgBuf() ;
  };

 #if (1)	/* compatibility with previous versions */
  typedef UmiMsgBuf UmiMsgDownBuf;
  typedef UmiMsgBuf UmiMsgUpBuf;
 #endif

  /**
   * user/driver copy
   */
  void copy(UmiData *, const UmiMsgBuf*);
  void copy(UmiMsgBuf*, const UmiData*);

  class UmiChanKern;
  /**
   * upper message handler interface definition
   */
  class UmiTimerSched {
   public:

    /*
     *  tiNext() schedules a call to 
     *  not earlier than the "current time" + tiDelta (the target time).
     *
     *  It returns the target time, which must be
     *  used by timer() to qualify the time at call as
     *  being equal or later than than the target time.
     *
     *  tiDelta is limited to tiLim(). 
     *  
     */
    virtual Time tiNext( const Time &tiDelta )=0;		

    UmiTimerSched() {}
    virtual ~UmiTimerSched() {}
  };

  class UmiMsgSig {
   public:

    /**
     * "get ready" signal function
     *
     * sigGet() is called by the driver when there is a
     * new message ready for pickup at the upstream queue.
     * The derived class may use the signal to continue
     * any thread waiting for upstream data.
     * The default implementation of sigGet() does nothing.
     *
     * \arg puc is the channel getting ready.
     * You can set and use an enviroment specific handle
     * for each channel via UmiChanKern::setHandle(void *p). Alternatively,
     * UmiChanKern::iChan() may be used to fetch the channel's number (starting at 0 for the first channel).
     *
     * \arg iCtx is the context identifier (i.e. the iCtx argument of the top level function resulting
     * in a call to setGet()).
     */
    virtual void sigGet( int iCtx, UmiChanKern *puc )=0;		

    /**
     * "put ready" signal function
     *
     * sigPut() is called by the driver when there is room
     * at the downstream queue to accommodate at least one
     * downstream message. 
     * The default implementation of sigPut() does nothing.
     * For argument descriptions see sigGet().
     */
    virtual void sigPut( int iCtx, UmiChanKern *puc )=0;		

    UmiMsgSig() {}
    virtual ~UmiMsgSig() {}
  };

  class UmiEnv: public UmiTimerSched, public UmiMsgSig {
   public:
    Time tiNext( const Time &tiDelta )=0;
    void sigGet( int iCtx, UmiChanKern *puc );		
    void sigPut( int iCtx, UmiChanKern *puc );		
    UmiEnv() {}
    virtual ~UmiEnv() {}
  };

  class UmiDriver;
  class UmiPlex;

  /**
   * Umi Channel - provides a transport independent interface to UmiPlexer.
   *
   */
  class UmiChannel : public Link {

    /* 
     * move messages from channel to channel
     */
    friend bool moveDownUp(UmiChannel *pucFrom, UmiChannel *pucTo);
    friend bool moveUpDown(UmiChannel *pucFrom, UmiChannel *pucTo);
   private:
    const UmiProto *_pup;
    UmiPlex *_pux;				/* multiplexer we're connected to */
    int _hChan;					/* handle @_pux */
    unsigned _cDriverMax;			/* max msgs concurrently scheduled to driver */
    
    /*
     * message pools
     */
    ChainRaw _cnUmuPool;
    ChainRaw _cnUmdPool;

    /*
     * downstream messages
     */
    ChainRaw _cnUmdQueue; 		
    ChainRaw _cnUmdDriver;
    ChainRaw _cnUmdUser;			

    /* 
     * upstream messages 
     */
    ChainRaw _cnUmuQueue;
    ChainRaw _cnUmuDriver;			
    ChainRaw _cnUmuUser;			

   protected:
    ChainRaw *pcnUmuPool() { return &_cnUmuPool; }
    ChainRaw *pcnUmuUser() { return &_cnUmuUser; }	/**< in process by the derived class */
    ChainRaw *pcnUmuDriver() { return &_cnUmuDriver; }	/**< in process by the driver/plexer */
    ChainRaw *pcnUmuQueue() { return &_cnUmuQueue; }	/**< msgs sent upstream (waiting)    */

    ChainRaw *pcnUmdPool() { return &_cnUmdPool; }
    ChainRaw *pcnUmdUser() { return &_cnUmdUser; }	/**< in process by the derived class */
    ChainRaw *pcnUmdDriver() { return &_cnUmdDriver; }  /**< in process by the driver/plexer */
    ChainRaw *pcnUmdQueue() { return &_cnUmdQueue; }    /**< downstream message waiting      */
    


   public:
    /* init buffers  */
    void initPoolDown(UmiMsgBuf *pUmd) { _cnUmdPool.push(pUmd); }
    void initPoolUp(UmiMsgBuf *pUmu) { _cnUmuPool.push(pUmu); }

    virtual void finiPool(Link *);      /**< release queue element, default implementation nops */


    /**
     * @name Implementation Interface
     */ //@{

    virtual void sigGet(int iCtx)=0;			
      /**< is called by the base class when there is a new 
           message ready for pickup at the upstream queue. */

    virtual void sigPut(int iCtx)=0;
     /**< is called by the base class when there is room at the 
          downstream queue to accommodate at least one downstream message.  */


    /**
     * Upstream message available if nonzero.
     */
    int pollIn()  { return pcnUmuQueue()->c(); } 
    UmiMsgBuf *getUpBuf();		/**< get from up queue, iff pollIn()	*/
    void releaseUpBuf(int iCtx, UmiMsgBuf *);	/**< release most recent got from getUpBuf()	*/

    /**
     * Space for at least one downstream message if nonzero.
     */
    int pollOut() { return pcnUmdPool()->c(); } 
    UmiMsgBuf* allocDownBuf();	/**< allocate a downstream buffer	*/
    void putDownBuf(int iCtx, UmiMsgBuf*);	/**< put downstream buffer to driver		*/

    //@}


    #ifndef _MM_NO_IOSTREAM
     std::ostream &printQueues(std::ostream &os) const ;
    #endif 

    void finiQueues();			/**< move all back to the pools and call finiPool()	*/
    void finiPools();			/**< call finiPool() for each queue element             */
    const UmiProto *pup() const { return _pup; }

    /**
     * set this channel's plexer port
     */
    void initPlex(UmiPlex *pux, int hChan, int cDriverMax=0);
    void finiPlex();
    UmiPlex *pux() const { return _pux; }		/**< this channels multiplexer	*/
    int hChan() const { return _hChan; }		/**< this channels handle \@pux()	*/
    

    /**
     * @name Plexer Interface
     */ //@{

    bool isUmuUser() { return (pcnUmuUser()->head() != 0); }

    /**
     * called by driver during sigDown(), since v130
     */
    Pmd getDown(int iCtx);
    Pmd peekDown() { return static_cast<UmiMsg*>( pcnUmdQueue()->head()); }

    /**
     * plexer's current downstream messages (imported by getDown())
     *
     * offset in order of the driver(plexer) queue, i.e. 
     *
     * 0 1st msg imported by getDown()
     * 1 2nd msg imported by getDown()
     * ...
     */
    Pmd _pmdDown(int iOffs);
    Pmd pmdDown(int iOffs=0) { return (iOffs) ? _pmdDown(iOffs) : static_cast<UmiMsg*>( pcnUmdDriver()->head()); }

    /**
     * plexer's current upstream messages (allocated by allocUp())
     *
     * offset in order of the driver(plexer) queue, i.e. 
     *
     * 0 1st msg allocated by allocUp()
     * 1 2nd msg allocated by allocUp()
     * ...
     */
    Pmu _pmuUp(int iOffs=0);
    Pmu pmuUp(int iOffs=0) { return (iOffs) ? _pmuUp(iOffs) : static_cast<UmiMsg*>( pcnUmuDriver()->head()); }


    /**
     * send a message upstream, message and buffer must have been obtained
     * by allocUp() and are released by this call. If fFlush is false, sigGet() is not called.
     * This is useful to fill the upstream queue in a loop without actually sending data. The last
     * call of putUp in the loop must set fFlush true.
     */
    void putUp(int iCtx, Pmu, bool fFlush=true );

    /**
     * allocate an upstream message
     */
    Pmu allocUp(size_t sData);

    /**
     * allocate an upstream message. If the queue is
     * blocked, remove the last message thas has been put upstream via putUp() and reallocate it.
     */
    Pmu allocUpForced(size_t sData);

    /**
     * allocate an upstream message. Declare that the control
     * and data buffers of pmd are no more used and may be
     * reused for the upstream message. 
     */
    Pmu allocUp(size_t sData, Pmd pmd);


    /**
     * release an downstream message, ANY downstream message must
     * be released this way. Clears channel. 
     * \note UmiDriver::putDown() MAY BE CALLED DURING releaseDown() FOR THE CHANNEL.
     */
    void releaseDown(int iCtx, Pmd);

    //@}

    /* data max per message */
    size_t sDataMax() { return __SZ_DATA_MAX; }		

    UmiChannel(const UmiProto *pup=0):_pup(pup) , _pux(0), _hChan(-1), _cDriverMax(0) {}	
    ~UmiChannel() {}
  };


  /**
   * kernel space channel implementation
   */
  class UmiChanKern : public UmiChannel {
    /* driver */
    UmiMsgSig*	_pue;

    /* channel number, identifies channel @ driver */
    void*	_phEnv;		/* handle at _pue		*/

   protected:
    void sigGet(int);
    void sigPut(int);
   public:

    UmiMsgSig *pue() { return _pue; }		/**< get the channels environment */



    /**
     * the the environment specific handle for this channel.
     * The default handle is 0.
     */
    void* getHandle() { return _phEnv; }

    /**
     * set the enviroment specific handle for this channel. 
     */
    void setHandle(void *p) { _phEnv=p; }

    UmiChanKern(UmiMsgSig *pue);
    virtual ~UmiChanKern();
  };

  /**
   * Multiplexer base class
   */
  class UmiPlex : public Link  {
   private:
    const char *_szName;
   public:
    /**
     * @name channel connection management 
     */ //@{

    /** 
     * open of a driver's channel - derived class 
     * \return Plexer handle 
     */
    virtual int openChannel( const char *szPort, UmiChannel *pud );

    /** 
     * close of a driver's channel - derived class
     */
    virtual void closeChannel( int iChan );
    //@}

    typedef int DownE;

    /**
     * @name pre-130 interface
     *
     * putDown() has been replaced by sigDown(),
     * backenable() replaced by sigUp()
     */
    //@{
    /** 
     * downstream message put for channel \ref iChan.
     * The buffer may be reused for
     * upstream message transfer, see UmiChannel::allocUp().
     * The message channel is locked until UmiChannel::releaseDown() is called.
     */ 
    virtual DownE  putDown(int iCtx, Pmd pmd, int iChan);

    /**
     * channel back-enable 
     */
    virtual void backenable(int iCtx, int iChan);
    //@}


    /**
     * @name signal interface, driven by channel
     *
     */
    //@{
    virtual void sigDown(int iCtx, UmiChannel *puc, int hChan);	/**< puc->getDown() ready  */
    virtual void sigUp(int iCtx, UmiChannel *puc, int hChan);	/**< puc->allocUp() ready  */
    //@}


    const char *szName() { return _szName ; }		/* plexer name */
    UmiPlex(const char *szName):  _szName(szName) { }
    virtual ~UmiPlex() { }
  };


  /**
   * timer interface def
   */
  class UmiCard;

  /**
   * lower message handler (driver) base class 
   */
  class UmiDriver : public UmiPlex {
   public:
   private:
    const char *_szName;
    UmiCard *_pca;
    ChainRaw _cnuc;			/* channel list		*/

   public:
    UmiChannel *pcn(int i) { return static_cast<UmiChannel*>(_cnuc.get(i)); }
    int ccn() { return _cnuc.c(); }

    void startIterChannels( ChainIterator *pit) { _cnuc.strtIter(pit); }
    UmiChannel *iterChannels( ChainIterator *pit) { return static_cast<UmiChannel*>(_cnuc.iter(pit)); }

    UmiCard *pca() const { return _pca; }



    /**
     * per channel (driver?) protocol descriptor
     */
    virtual const UmiProto *pup(int iChan)=0;


    int isOpen();	/**< is this driver open ? */

    /**
     * open of a driver - derived class
     *
     * Called after this driver has been made the active driver of the card.
     * Any derived function may enable interrupt sources. 
     */
    virtual int vinit();

    /**
     * close of a driver - derived class
     *
     * Any derived function must disable interrupt sources.
     */
    virtual void vfini();

    virtual char *print( char *, int ) const;

    /** interupt handler */
    virtual int interrupt(int iCtx)=0;

    /** timer handler */
    virtual int timer(int iCtx, const Time &)=0;

    /** max size of message data part */
    size_t sizeDataMax() const { return __SZ_DATA_MAX; }


    void init(UmiChannel *pcn) { _cnuc.push(pcn); }
    void fini(UmiChannel *pcn) { _cnuc.remove(pcn); }


    int init();		/**< activate driver, make this driver the active driver for the card. */
    void fini();	/**< idle driver, another driver may init() to access the card.     */

    /**
     * the driver is permanently bound to pca.
     * Multiple driver may share a pca. 
     */
    UmiDriver(const char *szName, UmiCard *pca);
    virtual ~UmiDriver();
  };


  /**
   * Card abstraction
   */
  class UmiCard {
    ChainRaw _cnud;		/* all drivers		*/
    UmiDriver *_pudCur;		/* "current" driver	*/
    UmiTimerSched *_pts;

   protected:


    /* to be called exclusively from UmiDriver::init()/fini() */
    friend class UmiDriver;
    int setDriver(UmiDriver *pud) { if (_pudCur) return -1; _pudCur=pud; return 0; }
    int clrDriver(UmiDriver *pud) { if (_pudCur != pud) return -1; _pudCur=0; return 0; }

   public:
    /** 
     * next timer tick >= now + ti (depends on system clock granularity)
     */
    Time tiNext( const Time &ti ) { return _pts->tiNext(ti); }

    UmiDriver *pudCur() { return _pudCur; }


    void startIterDrivers( ChainIterator *pit) { _cnud.strtIter(pit); }
    UmiDriver *iterDrivers( ChainIterator *pit) { return static_cast<UmiDriver*>(_cnud.iter(pit)); }



    /** 
     * per card primary interrupt function.
     * To be called from the enviroment whenever an interrupt from this
     * card might have been occured. As interrupts may be shared, a call
     * of interrupt() might have been done for another card.
     */
    virtual int interrupt(int iCtx)=0;

    /** per card "hard" timer */
    virtual int timer(int iCtx, const Time &)=0;

    virtual char *print( char *, int ) const ;

    void initDriver(UmiDriver *pud) { _cnud.push(pud); }
    UmiCard(UmiTimerSched *pts):_pudCur(0), _pts(pts) {}
    virtual ~UmiCard() {}
  };

  /* new namespace specific type names */
  typedef UmiPlex Plexer; 
  typedef UmiChannel Channel; 
  typedef UmiProto Proto; 
  typedef UmiChanKern ChanKern;

}  


#endif


