#ifndef _UMIMOD_H_
#define _UMIMOD_H_

#include <linux/module.h>
#include <linux/pci.h>
#include <linux/usb.h>
#include <linux/poll.h>
#include "umic/umiad.h"

  /* mapping of minor numbers to card/chan/driver/chan-of-driver */
  #define MAX_CNCC 8		/* max access controllers per chan	*/
  #define MAX_CACN 16		/* max chan per card			*/
  #define MAX_DRIV 4		/* max drivers per card			*/
  #define MAX_DRCN 4		/* max channels per driver		*/

  #define GETCARD(i) ((i) >> 4)		/* card index			*/
  #define GETCHAN(i) ((i) & 0xf)	/* straight channel 		*/
  #define GETDRIV(i) (((i) & 0xc) >> 2)	/* driver component		*/
  #define GETDRCN(i) (((i) & 0x3))	/* chan-of-driver comp		*/
  
  #define MAX_CARD (16)
  #define SZ_CARD_ID 64

  struct CardStruct;

  #define PFX "%s: "	/* kernel log prefix */


  /*
   * structure identifying the Linux UMI module to the UMI core
   */
  struct LumiModStruct {
    char *_szDid;		/* driver short name		*/
    const char *_szName;	/* driver long name		*/
    const char *_szDvs;		/* driver version		*/

    struct UmiMod *_pum;	/* per card UmiMod		*/
    int * _pfDebug;		/* pointer to debug flag	*/
    struct file_operations *_pFops;


    /* this modules ISA resources */
    int _cMaxIsa;		/* max isa cards		*/
    int *_piIoIsa;		/* ports array			*/
    int *_piIrqIsa;		/* irq array			*/
    int *_piTypeIsa;		/* type array			*/
    const UmiAdapt **_ppad;     /* all adaptor descriptors == _pum->_mppad */

    struct isa_driver *_pIdr;   /* for ISA cards		*/
    int *_pcIsa;		/* number of ISA devices	*/
    struct pci_driver *_pPdr;   /* if PCI supported 		*/
    struct usb_driver *_pUdr;	/* if USB supported		*/
    char *_szProcDir;		/* subdir of /proc/		*/

    /* 
     * default initialization to zero from here 
     */
    int _iMajor;		/* assigned major number	*/
    struct class *_pcl;		/* interface class		*/


    /*
     * cards allocated and 
     * managed by the specific bus modules
     */
    struct RegCards {
      struct CardStruct *_pca;		/* allocated and managed by bus module	*/
      char _mzId[SZ_CARD_ID];		/* identifies card physically		*/
    } _mprc[MAX_CARD];
  };

  typedef struct LumiModStruct LumiMod ;

  /* module debug */
  #define isDebugMod(p) (*((p)->_pfDebug))

  /* per card debug, to activate set bit (1<<card_no) in fDebug */
  #define isDebug(p, c) (*((p)->_pfDebug) && pdm_isDebug(p, c))
  #define plm_pum(p) ((p)->_pum)


  /*
   * allocate a slot.
   *
   * if sz is known, that slot is returned.
   * else a new slot is allocated and assigned to sz
   *
   * once assigned, slots can be released only by reset.
   */
  extern int pdm_allocslot(LumiMod *plm, const char *sz);
  extern int pdm_findslot(LumiMod *pdm, const char *sz);
  extern int pdm_getslot(LumiMod *pdm, const char *sz);
  extern struct CardStruct* pdm_setcard(LumiMod *plm, int i, struct CardStruct *pca);
  extern struct CardStruct* pdm_clrcard(LumiMod *plm, int i);
  extern struct CardStruct* pdm_getcard(LumiMod *plm, int i);
  extern char* pdm_getid(LumiMod *plm, struct CardStruct *pca);
  extern int pdm_isDebug(LumiMod *plm, struct CardStruct *pca);



  /* file operations */
  struct inode;
  struct file;
  extern int umi_open(LumiMod *pdm, struct inode * pin, struct file *fp) ;
  extern int umi_release(struct inode * pin, struct file *fp);
  extern int umi_ioctl(struct inode * pin,struct file *fp, unsigned int cmd, unsigned long arg );
  extern long umi_unlocked_ioctl(struct file *fp, unsigned int cmd, unsigned long arg );
  extern unsigned int umi_poll(struct file *fp, poll_table *pt );

  /* 
   * module_init(_exit) of UMI module  
   *
   * creates (destroys) class according to plm->_szDid
   * (un-)registers with the card specific cores (usb/pci/isa/...)
   */
  int pdm_init_module(LumiMod *pdm); 	
  void pdm_cleanup_module(LumiMod *pdm);


  /* enumerator for ISA */
  int enum_isa_devices(LumiMod *);
  int free_isa_devices(LumiMod *);

  /* isa */
  extern int pdm_isa_probe(LumiMod*, struct device *pdv, unsigned n);
  extern int pdm_isa_remove(LumiMod*, struct device *pdv, unsigned n);

  /* pci */
  extern int pdm_pci_probe(LumiMod*, struct pci_dev *pdv, const struct pci_device_id *pdi);
  extern void pdm_pci_remove(LumiMod*, struct pci_dev *pdv);

  /* usb */
  extern int pdm_usb_probe(LumiMod*, struct usb_interface *pui, const struct usb_device_id *pdi);
  extern void pdm_usb_disconnect(LumiMod*, struct usb_interface *pui);


#endif
